@@ -122,6 +122,7 @@ index((eval_license), 0, backward_compatibility, backcomp).
 index((trie_insert_new), 3, backward_compatibility, backcomp).
 index((thread_at_exit), 1, backward_compatibility, backcomp).
 index((read_history), 6, backward_compatibility, backcomp).
+index(('$sig_atomic'), 1, backward_compatibility, backcomp).
 index((base32), 2, base32, base32).
 index((base32), 3, base32, base32).
 index((base64_encoded), 3, base64, base64).
@@ -507,6 +508,8 @@ index((subset), 2, lists, lists).
 index((subtract), 3, lists, lists).
 index((main), 0, prolog_main, main).
 index((argv_options), 3, prolog_main, main).
+index((argv_options), 4, prolog_main, main).
+index((argv_usage), 1, prolog_main, main).
 index((cli_parse_debug_options), 2, prolog_main, main).
 index((cli_enable_development_system), 0, prolog_main, main).
 index((make), 0, make, make).
@@ -534,6 +537,7 @@ index((current_temporary_module), 1, modules, modules).
 index((mqi_start), 0, mqi, mqi).
 index((mqi_start), 1, mqi, mqi).
 index((mqi_stop), 1, mqi, mqi).
+index((mqi_version), 2, mqi, mqi).
 index((nb_rb_insert), 3, nb_rbtrees, nb_rbtrees).
 index((nb_rb_get_node), 3, nb_rbtrees, nb_rbtrees).
 index((nb_rb_node_value), 2, nb_rbtrees, nb_rbtrees).
@@ -822,6 +826,7 @@ index((trace), 1, prolog_trace, prolog_trace).
 index((trace), 2, prolog_trace, prolog_trace).
 index((tracing), 2, prolog_trace, prolog_trace).
 index((notraceall), 0, prolog_trace, prolog_trace).
+index((require_prolog_version), 2, prolog_versions, prolog_versions).
 index((wrap_predicate), 4, prolog_wrap, prolog_wrap).
 index((unwrap_predicate), 2, prolog_wrap, prolog_wrap).
 index((current_predicate_wrapper), 4, prolog_wrap, prolog_wrap).
@@ -932,9 +937,11 @@ index((maybe), 2, random, random).
 index((random_perm2), 4, random, random).
 index((random_member), 2, random, random).
 index((random_select), 3, random, random).
+index((random_subseq), 3, random, random).
 index((randseq), 3, random, random).
 index((randset), 3, random, random).
 index((random_permutation), 2, random, random).
+index((random_numlist), 4, random, random).
 index((random), 3, random, random).
 index((rb_new), 1, rbtrees, rbtrees).
 index((rb_empty), 1, rbtrees, rbtrees).
@@ -1281,6 +1288,9 @@ index((term_subsumer), 3, terms, terms).
 index((term_factorized), 3, terms, terms).
 index((mapargs), 3, terms, terms).
 index((mapsubterms), 3, terms, terms).
+index((mapsubterms_var), 3, terms, terms).
+index((foldsubterms), 4, terms, terms).
+index((foldsubterms), 5, terms, terms).
 index((same_functor), 2, terms, terms).
 index((same_functor), 3, terms, terms).
 index((same_functor), 4, terms, terms).
@@ -1468,6 +1478,9 @@ index((/), 9, yall, yall).
 index((lambda_calls), 2, yall, yall).
 index((lambda_calls), 3, yall, yall).
 index((is_lambda), 1, yall, yall).
+index((yaml_read), 2, yaml, yaml).
+index((yaml_write), 2, yaml, yaml).
+index((yaml_write), 3, yaml, yaml).
 index((zip_open), 4, zip, zip).
 index((zip_close), 1, zip, zip).
 index((zip_close), 2, zip, zip).
