@@ -3,7 +3,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2014, VU University Amsterdam
+    Copyright (c)  2014-2022, VU University Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -39,9 +40,12 @@
 :- use_module(library(option)).
 :- use_module(library(error)).
 :- use_module(library(debug)).
+:- use_module(library(http/json)).
 
 :- multifile
-    blob_rendering//3.              % +Type, +Blob, +Options
+    blob_rendering//3,              % +Type, +Blob, +Options
+    portray//2,                     % +Term, +Options
+    layout/3.                       % +Term, -Layout, +Options
 
 /** <module> Represent Prolog terms as HTML
 
@@ -79,20 +83,30 @@ term(Term, Options) -->
                     Options1),
       dict_options(Dict, Options1)
     },
-    any(Term, Dict).
+    any(Term, Dict),
+    finalize_term(Term, Dict).
 
+:- html_meta
+    embrace(html,?,?).
 
 any(_, Options) -->
     { Options.depth >= Options.max_depth },
     !,
     html(span(class('pl-ellipsis'), ...)).
+any(Term, Options) -->
+    (   {   nonvar(Term)
+        ;   attvar(Term)
+        }
+    ->  portray(Term, Options)
+    ),
+    !.
 any(Term, Options) -->
     { primitive(Term, Class0),
       !,
       quote_atomic(Term, S, Options),
       primitive_class(Class0, Term, S, Class)
     },
-    html(span(class(Class), S)).
+    html(span([class(Class)], S)).
 any(Term, Options) -->
     { blob(Term,Type), Term \== [] },
     !,
@@ -122,7 +136,7 @@ compound('$VAR'(Var), Options) -->
       ;   Class = 'pl-var'
       )
     },
-    html(span(class(Class), S)).
+    html(span([class(Class)], S)).
 compound(List, Options) -->
     { (   List == []
       ;   List = [_|_]                              % May have unbound tail
@@ -144,28 +158,45 @@ compound(OpTerm, Options) -->
     op1(Type, Pri, OpTerm, ArgPri, Options).
 compound(OpTerm, Options) -->
     { compound_name_arity(OpTerm, Name, 2),
-      is_op2(Name, LeftPri, Pri, RightPri, Options),
+      is_op2(Name, Type, LeftPri, Pri, RightPri, Options),
       \+ Options.get(ignore_ops) == true
     },
     !,
-    op2(Pri, OpTerm, LeftPri, RightPri, Options).
+    op2(Pri, OpTerm, Type, LeftPri, RightPri, Options).
 compound(Compound, Options) -->
     { compound_name_arity(Compound, Name, Arity),
       quote_atomic(Name, S, Options.put(embrace, never)),
       arg_options(Options, _{priority:999}, ArgOptions),
-      extra_classes(Classes, Options)
+      extra_classes(Compound, Classes, Attrs, Options)
     },
-    html(span(class(['pl-compound'|Classes]),
-              [ span(class('pl-functor'), S),
-                '(',
-                \args(0, Arity, Compound, ArgOptions),
-                ')'
+    html(span([ class(['pl-compound','pl-adaptive'|Classes]),
+                'data-arity'(Arity),
+                'data-name'(Name)
+              | Attrs
+              ],
+              [ span(class(['pl-functor', 'pl-trigger']),
+                     [ S, \punct('(') ]),
+                span(class('pl-compound-args'),
+                     [ \args(0, Arity, Compound, ArgOptions)
+                     ])
               ])).
 
-extra_classes(['pl-level-0'], Options) :-
-    Options.depth == 0,
-    !.
-extra_classes([], _).
+extra_classes(Term, Classes, OAttrs, Options) :-
+    findall(A, extra_attr(Term, A, Options), Attrs),
+    partition(is_class_attr, Attrs, CAttrs, OAttrs),
+    maplist(arg(1), CAttrs, Classes).
+
+is_class_attr(class(_)).
+
+extra_attr(_, class('pl-level-0'), Options) :-
+    Options.depth == 0.
+extra_attr(Term, 'data-layout'(Data), Options) :-
+    layout(Term, Layout, Options),
+    (   is_dict(Layout)
+    ->  atom_json_dict(Data, Layout, [])
+    ;   Data = Layout
+    ).
+
 
 %!  arg_options(+Options, -OptionsOut) is det.
 %!  arg_options(+Options, +Extra, -OptionsOut) is det.
@@ -186,21 +217,35 @@ args(I, Arity, Compound, ArgOptions) -->
     { NI is I + 1,
       arg(NI, Compound, Arg)
     },
-    any(Arg, ArgOptions),
     (   {NI == Arity}
-    ->  []
-    ;   html(', '),
+    ->  html([ span(class('pl-compound-arg'), \any(Arg, ArgOptions)),
+               span(class(['pl-compound-close', 'pl-punct']), ')')
+             ])
+    ;   html(span(class('pl-compound-arg'),
+                  [ \any(Arg, ArgOptions), \punct(',') ])),
         args(NI, Arity, Compound, ArgOptions)
     ).
 
+punct(Punct) -->
+    html(span(class('pl-punct'), Punct)).
+
 %!  list(+List, +Options)//
 %
 %   Emit a list.  The List may have an unbound tail.
 
 list(List, Options) -->
-    html(span(class('pl-list'),
-              ['[', \list_content(List, Options),
-               ']'
+    { '$skip_list'(Length, List, Tail),
+      (   Tail == []
+      ->  Attr = ['data-length'(Length)]
+      ;   Attr = ['data-length'(Length), 'data-partial'(true)]
+      )
+    },
+    html(span([ class(['pl-list','pl-adaptive'])
+              | Attr
+              ],
+              [ span(class(['pl-list-open', 'pl-trigger', 'pl-punct']), '['),
+                \list_content(List, Options),
+                span(class(['pl-list-close', 'pl-punct']), ']')
               ])).
 
 list_content([], _Options) -->
@@ -208,26 +253,36 @@ list_content([], _Options) -->
     [].
 list_content([H|T], Options) -->
     !,
-    { arg_options(Options, ArgOptions)
+    { arg_options(Options, ArgOptions),
+      (   T == []
+      ->  Sep = [],
+          Next = end
+      ;   Options.depth + 1 >= Options.max_depth
+      ->  Sep = [span(class('pl-punct'), '|')],
+          Next = depth_limit
+      ;   (var(T) ; \+ T = [_|_])
+      ->  Sep = [span(class('pl-punct'), '|')],
+          Next = tail
+      ;   Sep = [span(class('pl-punct'), [',', ' '])],
+          Next = list
+      )
     },
-    any(H, Options),
-    (   {T == []}
-    ->  []
-    ;   { Options.depth + 1 >= Options.max_depth }
-    ->  html(['|',span(class('pl-ellipsis'), ...)])
-    ;   {var(T) ; \+ T = [_|_]}
-    ->  html('|'),
-        tail(T, ArgOptions)
-    ;   html(', '),
-        list_content(T, ArgOptions)
-    ).
+    html(span(class('pl-list-el'),
+              [ \any(H, Options) | Sep ])),
+    list_next(Next, T, ArgOptions).
 
-tail(Value, Options) -->
+list_next(end, _, _) --> !.
+list_next(depth_limit, _, _) -->
+    !,
+    html(span(class('pl-ellipsis'), ...)).
+list_next(tail, Value, Options) -->
     {   var(Value)
     ->  Class = 'pl-var-tail'
     ;   Class = 'pl-nonvar-tail'
     },
     html(span(class(Class), \any(Value, Options))).
+list_next(list, Tail, Options) -->
+    list_content(Tail, Options).
 
 %!  is_op1(+Name, -Type, -Priority, -ArgPriority, +Options) is semidet.
 %
@@ -244,11 +299,11 @@ argpri(fy, prefix,  Pri,  Pri).
 argpri(xf, postfix, Pri0, Pri) :- Pri is Pri0 - 1.
 argpri(yf, postfix, Pri,  Pri).
 
-%!  is_op2(+Name, -LeftPri, -Pri, -RightPri, +Options) is semidet.
+% ! is_op2(+Name, -Type, -LeftPri, -Pri, -RightPri, +Options) is semidet.
 %
 %   True if Name is an operator taking two arguments of Type.
 
-is_op2(Name, LeftPri, Pri, RightPri, Options) :-
+is_op2(Name, Type, LeftPri, Pri, RightPri, Options) :-
     operator_module(Module, Options),
     current_op(Pri, Type, Module:Name),
     infix_argpri(Type, LeftPri, Pri, RightPri),
@@ -273,7 +328,7 @@ operator_module(TypeIn, _) :-
 op1(Type, Pri, Term, ArgPri, Options) -->
     { Pri > Options.priority },
     !,
-    html(['(', \op1(Type, Term, ArgPri, Options), ')']).
+    embrace(\op1(Type, Term, ArgPri, Options)).
 op1(Type, _, Term, ArgPri, Options) -->
     op1(Type, Term, ArgPri, Options).
 
@@ -283,12 +338,16 @@ op1(prefix, Term, ArgPri, Options) -->
       FuncOptions = DepthOptions.put(embrace, never),
       ArgOptions  = DepthOptions.put(priority, ArgPri),
       quote_atomic(Functor, S, FuncOptions),
-      extra_classes(Classes, Options)
+      extra_classes(Term, Classes, Attrs, Options.put(op, prefix))
     },
-    html(span(class(['pl-compound'|Classes]),
-              [ span(class('pl-prefix'), S),
-                \space(Functor, Arg, FuncOptions, ArgOptions),
-                \any(Arg, ArgOptions)
+    html(span([ class(['pl-compound', 'pl-op', 'pl-prefix-op'|Classes]),
+                'data-arity'(1),
+                'data-name'(Functor)
+              | Attrs
+              ],
+              [ span(class('pl-functor'), S),
+                \space(Functor, Arg, o, a, FuncOptions, ArgOptions),
+                \op_arg(Arg, ArgOptions)
               ])).
 op1(postfix, Term, ArgPri, Options) -->
     { Term =.. [Functor,Arg],
@@ -296,66 +355,138 @@ op1(postfix, Term, ArgPri, Options) -->
       ArgOptions = DepthOptions.put(priority, ArgPri),
       FuncOptions = DepthOptions.put(embrace, never),
       quote_atomic(Functor, S, FuncOptions),
-      extra_classes(Classes, Options)
+      extra_classes(Term, Classes, Attrs, Options.put(op, postfix))
     },
-    html(span(class(['pl-compound'|Classes]),
-              [ \any(Arg, ArgOptions),
-                \space(Arg, Functor, ArgOptions, FuncOptions),
-                span(class('pl-postfix'), S)
+    html(span([ class(['pl-compound', 'pl-op', 'pl-postfix-op'|Classes]),
+                'data-arity'(1),
+                'data-name'(Functor)
+              | Attrs
+              ],
+              [ \op_arg(Arg, ArgOptions),
+                \space(Arg, Functor, a, o, ArgOptions, FuncOptions),
+                span(class('pl-functor'), S)
               ])).
 
-%!  op2(+Pri, +Term, +LeftPri, +RightPri, +Options)// is det.
+%!  op2(+Pri, +Term, +Type, +LeftPri, +RightPri, +Options)// is det.
 
-op2(Pri, Term, LeftPri, RightPri, Options) -->
+op2(Pri, Term, Type, LeftPri, RightPri, Options) -->
     { Pri > Options.priority },
     !,
-    html(['(', \op2(Term, LeftPri, RightPri, Options), ')']).
-op2(_, Term, LeftPri, RightPri, Options) -->
-    op2(Term, LeftPri, RightPri, Options).
+    embrace(\op2(Term, Type, LeftPri, RightPri, Options)).
+op2(_, Term, Type, LeftPri, RightPri, Options) -->
+    op2(Term, Type, LeftPri, RightPri, Options).
 
-op2(Term, LeftPri, RightPri, Options) -->
+op2(Term, xfy, LeftPri, RightPri, Options) -->
+    { functor(Term, Functor, 2),
+      quote_op(Functor, S, Options),
+      xfy_list(Term, Functor, List),
+      arg_options(Options, DepthOptions),
+      ArgOptions  = DepthOptions.put(#{priority:LeftPri, quoted_op:S}),
+      extra_classes(Term, Classes, Attrs, Options.put(op, infix))
+    },
+    html(span([ class(['pl-op-seq', 'pl-adaptive'|Classes])
+              | Attrs
+              ],
+              \op_seq(List, Functor, RightPri, ArgOptions))).
+op2(Term, _Type, LeftPri, RightPri, Options) -->
     { Term =.. [Functor,Left,Right],
       arg_options(Options, DepthOptions),
       LeftOptions  = DepthOptions.put(priority, LeftPri),
       FuncOptions  = DepthOptions.put(embrace, never),
       RightOptions = DepthOptions.put(priority, RightPri),
-      (   (   need_space(Left, Functor, LeftOptions, FuncOptions)
-          ;   need_space(Functor, Right, FuncOptions, RightOptions)
+      (   (   need_space(Left, Functor, a, o, LeftOptions, FuncOptions)
+          ;   need_space(Functor, Right, o, a, FuncOptions, RightOptions)
           )
       ->  Space = ' '
       ;   Space = ''
       ),
       quote_op(Functor, S, Options),
-      extra_classes(Classes, Options)
+      extra_classes(Term, Classes, Attrs, Options.put(op, infix))
     },
-    html(span(class(['pl-compound'|Classes]),
-              [ \any(Left, LeftOptions),
+    html(span([ class(['pl-compound', 'pl-op', 'pl-infix-op'|Classes]),
+                'data-arity'(2),
+                'data-name'(Functor)
+              | Attrs
+              ],
+              [ \op_arg(Left, LeftOptions),
                 Space,
-                span(class('pl-infix'), S),
+                span(class('pl-functor'), S),
                 Space,
-                \any(Right, RightOptions)
+                \op_arg(Right, RightOptions)
               ])).
 
-%!  space(@T1, @T2, +Options)//
+%!  op_arg(+Term, +Options)// is det.
+
+op_arg(Atom, Options) -->
+    { atom(Atom),
+      operator_module(Module, Options),
+      current_op(_,_,Module:Atom)
+    }, !,
+    embrace(\any(Atom, Options.put(embrace, never))).
+op_arg(Any, Options) -->
+    any(Any, Options).
+
+op_seq([Last], _Functor, LastPri, Options) -->
+    !,
+    { LastOptions = Options.put(priority, LastPri)
+    },
+    html(span(class('pl-op-seq-el'), \op_arg(Last, LastOptions))).
+op_seq([H|T], Functor, LastPri, Options) -->
+    html(span(class('pl-op-seq-el'),
+              [ \op_arg(H, Options),
+                \left_space(H, Functor, Options),
+                span(class('pl-infix'), Options.quoted_op)
+              ])),
+    op_seq(T, Functor, LastPri, Options).
+
+left_space(Left, Functor, Options) -->
+    { need_space(Left, Functor, a, o, Options, Options.put(embrace, never))
+    },
+    !,
+    html(' ').
+left_space(_,_,_) -->
+    [].
+
+xfy_list(Term, Name, List),
+    compound(Term),
+    compound_name_arguments(Term, _, [A,B]) =>
+    List = [A|T],
+    xfy_list(B, Name, T).
+xfy_list(Term, _, List) =>
+    List = [Term].
+
+%!  embrace(+HTML)//
+%
+%   Place parenthesis around HTML  with  a   DOM  that  allows to easily
+%   justify the height of the parenthesis.
+
+embrace(HTML) -->
+    html(span(class('pl-embrace'),
+              [ span(class('pl-parenthesis'), '('),
+                span(class('pl-embraced'),\html(HTML)),
+                span(class('pl-parenthesis'), ')')
+              ])).
+
+%!  space(@T1, @T2, +C1, +C2, +Options)//
 %
 %   Emit a space if omitting a space   between T1 and T2 would cause
 %   the two terms to join.
 
-space(T1, T2, LeftOptions, RightOptions) -->
-    { need_space(T1, T2, LeftOptions, RightOptions) },
+space(T1, T2, C1, C2, LeftOptions, RightOptions) -->
+    { need_space(T1, T2, C1, C2, LeftOptions, RightOptions) },
     html(' ').
-space(_, _, _, _) -->
+space(_, _, _, _, _, _) -->
     [].
 
-need_space(T1, T2, _, _) :-
+need_space(T1, T2, _, _, _, _) :-
     (   is_solo(T1)
     ;   is_solo(T2)
     ),
     !,
     fail.
-need_space(T1, T2, LeftOptions, RightOptions) :-
-    end_code_type(T1, TypeR, LeftOptions.put(side, right)),
-    end_code_type(T2, TypeL, RightOptions.put(side, left)),
+need_space(T1, T2, C1, C2, LeftOptions, RightOptions) :-
+    end_code_type(T1, C1, TypeR, LeftOptions.put(side, right)),
+    end_code_type(T2, C2, TypeL, RightOptions.put(side, left)),
     \+ no_space(TypeR, TypeL).
 
 no_space(punct, _).
@@ -366,11 +497,20 @@ no_space(quote(R), quote(L)) :-
 no_space(alnum, symbol).
 no_space(symbol, alnum).
 
-%!  end_code_type(+Term, -Code, Options)
+%!  end_code_type(+Term, +Class, -Code, Options)
 %
 %   True when code is the first/last character code that is emitted
 %   by printing Term using Options.
 
+end_code_type(Atom, a, Type, Options) :-
+    atom(Atom),
+    operator_module(Module, Options),
+    current_op(_,_,Module:Atom),
+    !,
+    Type = punct.
+end_code_type(Atom, _, Type, Options) :-
+    end_code_type(Atom, Type, Options).
+
 end_code_type(_, Type, Options) :-
     Options.depth >= Options.max_depth,
     !,
@@ -405,28 +545,38 @@ end_code_type(OpTerm, Type, Options) :-
     !,
     (   Pri > Options.priority
     ->  Type = punct
-    ;   (   OpType == prefix
+    ;   (   OpType == prefix, Options.side == left
+        ->  end_code_type(Name, Type, Options)
+        ;   OpType == postfix, Options.side == right
         ->  end_code_type(Name, Type, Options)
         ;   arg(1, OpTerm, Arg),
             arg_options(Options, ArgOptions),
-            end_code_type(Arg, Type, ArgOptions.put(priority, ArgPri))
+            op_end_code_type(Arg, Type, ArgOptions.put(priority, ArgPri))
         )
     ).
 end_code_type(OpTerm, Type, Options) :-
     compound_name_arity(OpTerm, Name, 2),
-    is_op2(Name, LeftPri, Pri, _RightPri, Options),
+    is_op2(Name, _Type, LeftPri, Pri, RightPri, Options),
     \+ Options.get(ignore_ops) == true,
     !,
     (   Pri > Options.priority
     ->  Type = punct
-    ;   arg(1, OpTerm, Arg),
+    ;   Options.side == left
+    ->  arg(1, OpTerm, Arg),
         arg_options(Options, ArgOptions),
-        end_code_type(Arg, Type, ArgOptions.put(priority, LeftPri))
+        op_end_code_type(Arg, Type, ArgOptions.put(priority, LeftPri))
+    ;   Options.side == right
+    ->  arg(2, OpTerm, Arg),
+        arg_options(Options, ArgOptions),
+        op_end_code_type(Arg, Type, ArgOptions.put(priority, RightPri))
     ).
 end_code_type(Compound, Type, Options) :-
     compound_name_arity(Compound, Name, _),
     end_code_type(Name, Type, Options).
 
+op_end_code_type(Atom, Type, Options) :-
+    end_code_type(Atom, a, Type, Options).
+
 end_type(S, Type, Options) :-
     number(S),
     !,
@@ -466,11 +616,14 @@ dict(Term, Options) -->
       quote_atomic(Tag, S, Options.put(embrace, never)),
       arg_options(Options, ArgOptions)
     },
-    html(span(class('pl-dict'),
-              [ span(class('pl-tag'), S),
-                '{',
-                \dict_kvs(Pairs, ArgOptions),
-                '}'
+    html(span(class(['pl-dict', 'pl-adaptive']),
+              [ span(class(['pl-tag', 'pl-trigger']), S),
+                span(class(['pl-dict-open', 'pl-punct']), '{'),
+                span(class('pl-dict-body'),
+                     [ span(class('pl-dict-kvs'),
+                            \dict_kvs(Pairs, ArgOptions)),
+                       span(class(['pl-dict-close', 'pl-punct']), '}')
+                     ])
               ])).
 
 dict_kvs([], _) --> [].
@@ -481,6 +634,8 @@ dict_kvs(_, Options) -->
 dict_kvs(KVs, Options) -->
     dict_kvs2(KVs, Options).
 
+dict_kvs2([], _) -->
+    [].
 dict_kvs2([K-V|T], Options) -->
     { quote_atomic(K, S, Options),
       end_code_type(V, VType, Options.put(side, left)),
@@ -488,18 +643,21 @@ dict_kvs2([K-V|T], Options) -->
       ->  VSpace = ' '
       ;   VSpace = ''
       ),
-      arg_options(Options, ArgOptions)
+      arg_options(Options, ArgOptions),
+      (   T == []
+      ->  Sep = []
+      ;   Sep = [\punct(','), ' ']
+      )
     },
-    html([ span(class('pl-key'), S),
-           ':',                             % FIXME: spacing
-           VSpace,
-           \any(V, ArgOptions)
-         ]),
-    (   {T==[]}
-    ->  []
-    ;   html(', '),
-        dict_kvs2(T, Options)
-    ).
+    html(span(class('pl-dict-kv'),
+              [ span(class('pl-key'), [S, \punct(:)]),
+                VSpace,
+                span(class('pl-dict-value'),
+                     [ \any(V, ArgOptions)
+                     | Sep
+                     ])
+              ])),
+    dict_kvs2(T, Options).
 
 quote_atomic(Float, String, Options) :-
     float(Float),
@@ -569,6 +727,23 @@ primitive_class('pl-atom', Atom, String, Class) :-
     Class = 'pl-quoted-atom'.
 primitive_class(Class, _, _, Class).
 
+%!  finalize_term(+Term, +Dict)// is det.
+%
+%   Handle the full_stop(Bool) and nl(Bool) options.
+
+finalize_term(Term, Dict) -->
+    (   { true == Dict.get(full_stop) }
+    ->  space(Term, '.', o, o, Dict, Dict),
+        (   { true == Dict.get(nl) }
+        ->  html(['.', br([])])
+        ;   html('. ')
+        )
+    ;   (   { true == Dict.get(nl) }
+        ->  html(br([]))
+        ;   []
+        )
+    ).
+
 
                  /*******************************
                  *             HOOKS            *
