@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2006-2017, University of Amsterdam
+    Copyright (c)  2006-2020, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -42,6 +43,7 @@
             print_html_head/1,          % +Stream
             predref//1,                 % +PI //
             predref//2,                 % +PI, Options //
+            nopredref//1,               % +PI //
             module_info/3,              % +File, +Options0, -Options
             doc_hide_private/3,         % +Doc0, -Doc, +Options
             edit_button//2,             % +File, +Options, //
@@ -70,6 +72,7 @@
             tags//1,                    % +Tags, //
             term//3,                    % +Text, +Term, +Bindings, //
             file_header//2,             % +File, +Options, //
+            flagref//1,                 % +Flag
             objects//2,                 % +Objects, +Options, //
             object_ref//2,              % +Object, +Options, //
             object_name//2,             % +Object, +Object
@@ -78,7 +81,9 @@
             object_page//2,             % +Object, +Options, //
             object_page_header//2,      % +File, +Options, //
             object_synopsis//2,         % +Object, +Options, //
-            object_page_footer//2       % +Object, +Options, //
+            object_footer//2,           % +Object, +Options, //
+            object_page_footer//2,      % +Object, +Options, //
+            cite//1                     % +Citations
           ]).
 :- use_module(library(lists)).
 :- use_module(library(option)).
@@ -102,8 +107,15 @@
 :- use_module(doc_search).
 :- use_module(doc_index).
 :- use_module(doc_util).
+:- use_module(library(solution_sequences)).
+:- use_module(library(error)).
+:- use_module(library(occurs)).
+:- use_module(library(prolog_source)).
+:- use_module(library(prolog_xref)).
+
 :- include(hooks).
 
+
 /** <module> PlDoc HTML backend
 
 This  module  translates  the  Herbrand   term  from  the  documentation
@@ -364,7 +376,17 @@ reply_file_objects(File, Objs0, Objects, FileOptions, Options) :-
     module_info(File, ModuleOptions, Options),
     file_info(Objs0, Objs1, FileOptions, ModuleOptions),
     doc_hide_private(Objs1, ObjectsSelf, ModuleOptions),
-    include_reexported(ObjectsSelf, Objects, File, FileOptions).
+    include_reexported(ObjectsSelf, Objects1, File, FileOptions),
+    remove_doc_duplicates(Objects1, Objects, []).
+
+remove_doc_duplicates([], [], _).
+remove_doc_duplicates([H|T0], [H|T], Seen) :-
+    H = doc(_, _, Comment),
+    \+ memberchk(Comment, Seen),
+    !,
+    remove_doc_duplicates(T0, T, [Comment|Seen]).
+remove_doc_duplicates([_|T0], T, Seen) :-
+    remove_doc_duplicates(T0, T, Seen).
 
 include_reexported(SelfObjects, Objects, File, Options) :-
     option(include_reexported(true), Options),
@@ -490,6 +512,8 @@ private(Object, _Options):-
     prolog:doc_is_public_object(Object), !, fail.
 private(Module:PI, Options) :-
     multifile(Module:PI, Options), !, fail.
+private(Module:PI, Options) :-
+    public(Module:PI, Options), !, fail.
 private(Module:PI, Options) :-
     option(module(Module), Options),
     option(public(Public), Options),
@@ -500,7 +524,7 @@ private(Module:PI, Options) :-
 private(Module:PI, _Options) :-
     module_property(Module, file(_)),      % A loaded module
     !,
-    export_list(Module, Exports),
+    module_property(Module, exports(Exports)),
     \+ ( member(PI2, Exports),
          eq_pi(PI, PI2)
        ).
@@ -523,7 +547,20 @@ multifile(Obj, _Options) :-
     pi_to_head(PI, Head),
     (   predicate_property(Module:Head, multifile)
     ;   xref_module(Source, Module),
-        xref_defined(Source, Head, multifile(_))
+        xref_defined(Source, Head, multifile(_Line))
+    ),
+    !.
+
+%!  public(+Options, +Options)
+%
+%   True if Obj is declared using public/1.
+
+public(Obj, _Options) :-
+    strip_module(user:Obj, Module, PI),
+    pi_to_head(PI, Head),
+    (   predicate_property(Module:Head, public)
+    ;   xref_module(Source, Module),
+        xref_defined(Source, Head, public(_Line))
     ),
     !.
 
@@ -768,7 +805,9 @@ pred_dom(Obj, Options, Pos-Comment, DOM) :-
             )
         ;   Class = multidef(file((Pos)))
         )
-    ;   Class = pubdef              % public definition
+    ;   public(Obj, Options)
+    ->  Class = publicdef           % :- public definition
+    ;   Class = pubdef              % exported definition
     ),
     (   Obj = Module:_
     ->  POptions = [module(Module)|Options]
@@ -1002,13 +1041,25 @@ file_link(File) -->
     html([ div(a(href(location_by_id(pldoc_doc)+File), File))
          ]).
 
+%!  object_footer(+Obj, +Options)// is det.
+%
+%   Call the hook prolog:doc_object_footer//2. This hook will be used to
+%   deal with examples.
+
+object_footer(Obj, Options) -->
+    prolog:doc_object_footer(Obj, Options),
+    !.
+object_footer(_, _) --> [].
+
+
 %!  object_page_footer(+Obj, +Options)// is det.
 %
 %   Call the hook prolog:doc_object_page_footer//2. This hook will
 %   be used to deal with annotations.
 
 object_page_footer(Obj, Options) -->
-    prolog:doc_object_page_footer(Obj, Options).
+    prolog:doc_object_page_footer(Obj, Options),
+    !.
 object_page_footer(_, _) --> [].
 
 
@@ -1033,10 +1084,17 @@ object_synopsis(Name/Arity, Options) -->
     object_synopsis(_:Name/Arity, Options).
 object_synopsis(M:Name/Arity, Options) -->
     { functor(Head, Name, Arity),
-      predicate_property(M:Head, exported),
-      \+ predicate_property(M:Head, imported_from(_)),
-      module_property(M, file(File)),
-      file_name_on_path(File, Spec),
+      (   option(source(Spec), Options)
+      ->  absolute_file_name(Spec, File,
+                             [ access(read),
+                               file_type(prolog),
+                               file_errors(fail)
+                             ])
+      ;   predicate_property(M:Head, exported),
+          \+ predicate_property(M:Head, imported_from(_)),
+          module_property(M, file(File)),
+          file_name_on_path(File, Spec)
+      ),
       !,
       unquote_filespec(Spec, Unquoted),
       (   predicate_property(Head, autoload(FileBase)),
@@ -1049,6 +1107,14 @@ object_synopsis(M:Name/Arity, Options) -->
     ->  synopsis([code([':- use_module(',a(href(HREF), '~q'-[Unquoted]),').'])|Extra])
     ;   synopsis([code(':- use_module(~q).'-[Unquoted])|Extra])
     ).
+object_synopsis(Name//Arity, Options) -->
+    !,
+    { DCGArity is Arity+2 },
+    object_synopsis(Name/DCGArity, Options).
+object_synopsis(Module:Name//Arity, Options) -->
+    !,
+    { DCGArity is Arity+2 },
+    object_synopsis(Module:Name/DCGArity, Options).
 object_synopsis(f(_/_), _) -->
     synopsis(span(class(function),
                   [ 'Arithmetic function (see ',
@@ -1056,7 +1122,8 @@ object_synopsis(f(_/_), _) -->
                     ')'
                   ])).
 object_synopsis(c(Func), _) -->
-    { sub_atom(Func, 0, _, _, 'PL_')
+    {   sub_atom(Func, 0, _, _, 'PL_')
+    ;   sub_atom(Func, 0, _, _, 'S')
     },
     !,
     synopsis([span(class(cfunc), 'C-language interface function')]).
@@ -1076,6 +1143,7 @@ synopsis(Text) -->
 unquote_filespec(Spec, Unquoted) :-
     compound(Spec),
     Spec =.. [Alias,Path],
+    atom(Path),
     atomic_list_concat(Parts, /, Path),
     maplist(need_no_quotes, Parts),
     !,
@@ -1635,6 +1703,13 @@ pred_head(Head) -->                     % Postfix operators
     html([ var(class=arglist, \pred_arg(Arg, 1)),
            ' ', b(class=pred, Functor)
          ]).
+pred_head({Head}) -->
+    !,
+    html([ b(class=pred, '{'),
+           var(class=arglist,
+               \pred_args([Head], 1)),
+           b(class=pred, '}')
+         ]).
 pred_head(Head) -->                     % Plain terms
     { Head =.. [Functor|Args] },
     html([ b(class=pred, Functor),
@@ -1845,6 +1920,45 @@ canonical_pi(Name//Arity, Name/Arity2, [Name, //, Arity]) :-
     !,
     Arity2 is Arity+2.
 
+%!  nopredref(+PI)//
+%
+%   Result of ``name/arity``, non-linking predicate indicator.
+
+nopredref(PI) -->
+    { canonical_pi(PI, _CPI, HTML)
+    },
+    !,
+    html(span(class=nopredref, HTML)).
+
+%!  flagref(+Flag)//
+%
+%   Reference to a Prolog flag.
+%
+%   @tbd generate a link to the Prolog website?
+
+flagref(Flag) -->
+    html(code(Flag)).
+
+%!  cite(+Citations)// is det.
+%
+%   Emit citations. This is indented to   allow  for [@cite1;@cite2] for
+%   generating LaTex.
+
+cite(Citations) -->
+    html('['), citations(Citations), html(']').
+
+citations([]) --> [].
+citations([H|T]) -->
+    citation(H),
+    (   {T==[]}
+    ->  []
+    ;   [';'],
+        citations(T)
+    ).
+
+citation(H) -->
+    html([@,H]).
+
 
 %!  manref(+NameArity, -HREF, +Options) is det.
 %
@@ -1926,8 +2040,8 @@ pred_source_href(Name/Arity, Module, HREF) :-
     (   catch(relative_file(Module:Head, File), _, fail)
     ->  uri_data(path, Components, File),
         uri_components(HREF, Components)
-    ;   in_file(Module:Head, File0),
-        insert_alias(File0, File),
+    ;   in_file(Module:Head, File0)
+    ->  insert_alias(File0, File),
         http_location_by_id(pldoc_doc, DocHandler),
         atom_concat(DocHandler, File, Path),
         uri_data(path, Components, Path),
@@ -2089,6 +2203,10 @@ object_name(_, file(File), _) -->
 object_name(_, directory(Dir), _) -->
     { file_base_name(Dir, Base) },
     html(Base).
+object_name(_, module(Title), _Options) -->
+    { print_message(warning,
+                    pldoc(module_comment_outside_module(Title)))
+    }.
 
 pi(title, PI, Options) -->
     pi_type(PI),
@@ -2129,28 +2247,58 @@ pi_type(_//_) -->
 
 in_file(Module:Head, File) :-
     !,
-    in_file(Module, Head, File).
+    distinct(File, in_file(Module, Head, File)).
 in_file(Head, File) :-
-    in_file(_, Head, File).
+    distinct(File, in_file(_, Head, File)).
 
-in_file(_, Head, File) :-
+in_file(Module, Head, File) :-
+    var(Module),
+    (   predicate_property(system:Head, foreign)
+    ->  !,
+        fail
+    ;   predicate_property(system:Head, file(File)),
+        \+ system_arithmetic_function(Head)
+    ->  !
+    ;   predicate_property(Head, autoload(File0))
+    ->  !,
+        file_name_extension(File0, pl, File)
+    ;   exported_from(Module, Head, File),
+        module_property(Module, class(library))
+    ).
+in_file(Module, Head, File) :-
+    xref_defined(File, Head, How),
     xref_current_source(File),
     atom(File),                     % only plain files
-    xref_defined(File, Head, How),
+    xref_module(File, Module),
     How \= imported(_From).
 in_file(Module, Head, File) :-
-    predicate_property(Module:Head, exported),
-    (   predicate_property(Module:Head, imported_from(Primary))
-    ->  true
-    ;   Primary = Module
-    ),
-    module_property(Primary, file(File)).
+    exported_from(Module, Head, File).
 in_file(Module, Head, File) :-
-    predicate_property(Module:Head, file(File)).
+    predicate_property(Module:Head, file(File)),
+    \+ predicate_property(Module:Head, imported_from(_)).
 in_file(Module, Head, File) :-
     current_module(Module),
     source_file(Module:Head, File).
 
+exported_from(Module, Head, File) :-
+    distinct(Primary,
+             (   predicate_property(Module:Head, exported),
+                 (   predicate_property(Module:Head, imported_from(Primary))
+                 ->  true
+                 ;   Primary = Module
+                 ))),
+    module_property(Primary, file(File)).
+
+:- multifile
+    arithmetic:evaluable/2.
+
+system_arithmetic_function(Head) :-
+    functor(Head, Name, Arity),
+    FArith is Arity-1,
+    FArith >= 0,
+    functor(FHead, Name, FArith),
+    arithmetic:evaluable(FHead, system).
+
 %%     file(+FileName)// is det.
 %%     file(+FileName, +Options)// is det.
 %
@@ -2502,3 +2650,8 @@ pred_anchor_name(//(Head), Name/Arity, Anchor) :-
 pred_anchor_name(Head, Name/Arity, Anchor) :-
     functor(Head, Name, Arity),
     format(atom(Anchor), '~w/~d', [Name, Arity]).
+
+:- multifile prolog:message//1.
+
+prolog:message(pldoc(module_comment_outside_module(Title))) -->
+    [ 'PlDoc comment <module> ~w does not appear in a module'-[Title] ].
