@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1995-2020, University of Amsterdam
+    Copyright (c)  1995-2021, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -43,6 +44,8 @@
 :- use_module(library(option)).
 :- use_module(library(error)).
 :- use_module(library(apply)).
+:- autoload(library(shlib), [current_foreign_library/2]).
+:- autoload(library(prolog_autoload), [autoload_all/1]).
 
 /** <module> Save current program as a state or executable
 
@@ -73,6 +76,8 @@ save_option(toplevel,    callable,
             "Toplevel goal").
 save_option(init_file,   atom,
             "Application init file").
+save_option(pce,         boolean,
+            "Do (not) include the xpce graphics subsystem").
 save_option(packs,       boolean,
             "Do (not) attach packs").
 save_option(class,       oneof([runtime,development]),
@@ -130,12 +135,11 @@ qsave_program(File) :-
     qsave_program(File, []).
 
 qsave_program(FileBase, Options0) :-
-    meta_options(is_meta, Options0, Options),
-    check_options(Options),
-    exe_file(FileBase, File, Options),
-    option(class(SaveClass),    Options, runtime),
-    option(init_file(InitFile), Options, DefInit),
-    default_init_file(SaveClass, DefInit),
+    meta_options(is_meta, Options0, Options1),
+    check_options(Options1),
+    exe_file(FileBase, File, Options1),
+    option(class(SaveClass), Options1, runtime),
+    qsave_init_file_option(SaveClass, Options1, Options),
     prepare_entry_points(Options),
     save_autoload(Options),
     setup_call_cleanup(
@@ -147,7 +151,7 @@ qsave_program(FileBase, Options0) :-
                                      % running on this state
           setup_call_catcher_cleanup(
               open(File, write, StateOut, [type(binary)]),
-              write_state(StateOut, SaveClass, InitFile, Options),
+              write_state(StateOut, SaveClass, Options),
               Reason,
               finalize_state(Reason, StateOut, File))
         ),
@@ -155,19 +159,16 @@ qsave_program(FileBase, Options0) :-
     cleanup,
     !.
 
-write_state(StateOut, SaveClass, InitFile, Options) :-
+write_state(StateOut, SaveClass, Options) :-
     make_header(StateOut, SaveClass, Options),
     setup_call_cleanup(
         zip_open_stream(StateOut, RC, []),
-        write_zip_state(RC, SaveClass, InitFile, Options),
+        write_zip_state(RC, SaveClass, Options),
         zip_close(RC, [comment('SWI-Prolog saved state')])),
     flush_output(StateOut).
 
-write_zip_state(RC, SaveClass, InitFile, Options) :-
-    save_options(RC, SaveClass,
-                 [ init_file(InitFile)
-                 | Options
-                 ]),
+write_zip_state(RC, SaveClass, Options) :-
+    save_options(RC, SaveClass, Options),
     save_resources(RC, SaveClass),
     lock_files(SaveClass),
     save_program(RC, SaveClass, Options),
@@ -199,16 +200,19 @@ exe_file(Base, Exe, Options) :-
     file_name_extension(Base, exe, Exe).
 exe_file(Exe, Exe, _).
 
-default_init_file(runtime, none) :- !.
-default_init_file(_,       InitFile) :-
-    '$cmd_option_val'(init_file, InitFile).
-
 delete_if_exists(File) :-
     (   exists_file(File)
     ->  delete_file(File)
     ;   true
     ).
 
+qsave_init_file_option(runtime, Options1, Options) :-
+    \+ option(init_file(_), Options1),
+    !,
+    Options = [init_file(none)|Options1].
+qsave_init_file_option(_, Options, Options).
+
+
                  /*******************************
                  *           HEADER             *
                  *******************************/
@@ -285,12 +289,12 @@ doption(nosignals).
 save_options(RC, SaveClass, Options) :-
     zipper_open_new_file_in_zip(RC, '$prolog/options.txt', Fd, []),
     (   doption(OptionName),
-            '$cmd_option_val'(OptionName, OptionVal0),
-            save_option_value(SaveClass, OptionName, OptionVal0, OptionVal1),
-            OptTerm =.. [OptionName,OptionVal2],
-            (   option(OptTerm, Options)
+            (   OptTerm =.. [OptionName,OptionVal2],
+                option(OptTerm, Options)
             ->  convert_option(OptionName, OptionVal2, OptionVal, FmtVal)
-            ;   OptionVal = OptionVal1,
+            ;   '$cmd_option_val'(OptionName, OptionVal0),
+                save_option_value(SaveClass, OptionName, OptionVal0, OptionVal1),
+                OptionVal = OptionVal1,
                 FmtVal = '~w'
             ),
             atomics_to_string(['~w=', FmtVal, '~n'], Fmt),
@@ -1003,7 +1007,7 @@ strip_file(File, Stripped) :-
 strip_file(File, File).
 
 do_strip_file(Strip, File, Stripped) :-
-    format(atom(Cmd), '"~w" -o "~w" "~w"',
+    format(atom(Cmd), '"~w" -x -o "~w" "~w"',
            [Strip, Stripped, File]),
     shell(Cmd),
     exists_file(Stripped).
